package ru.dbotthepony.kstarbound.io

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import com.google.gson.GsonBuilder
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.internal.bind.TypeAdapters
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import it.unimi.dsi.fastutil.objects.Object2IntArrayMap
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kstarbound.Starbound
import java.lang.reflect.Constructor
import kotlin.jvm.internal.DefaultConstructorMarker
import kotlin.reflect.*
import kotlin.reflect.full.isSuperclassOf
import kotlin.reflect.full.isSupertypeOf
import kotlin.reflect.full.memberProperties

@Suppress("unchecked_cast")
private fun <T> resolveBound(bound: Class<T>): TypeAdapter<T> {
	return when (bound) {
		Float::class.java -> TypeAdapters.FLOAT as TypeAdapter<T>
		Double::class.java -> TypeAdapters.DOUBLE as TypeAdapter<T>
		String::class.java -> TypeAdapters.STRING as TypeAdapter<T>
		Int::class.java -> TypeAdapters.INTEGER as TypeAdapter<T>
		Long::class.java -> TypeAdapters.LONG as TypeAdapter<T>
		Boolean::class.java -> TypeAdapters.BOOLEAN as TypeAdapter<T>
		else -> Starbound.gson.getAdapter(bound)
	}
}

class ListAdapter<T>(private val bound: Class<T>) : TypeAdapter<List<T>>() {
	private val resolvedBound by lazy {
		resolveBound(bound)
	}

	override fun write(out: JsonWriter, value: List<T>) {
		out.beginArray()

		val resolvedBound = resolvedBound

		for (v in value) {
			resolvedBound.write(out, v)
		}

		out.endArray()
	}

	override fun read(reader: JsonReader): List<T> {
		reader.beginArray()

		val builder = ImmutableList.builder<T>()
		val resolvedBound = resolvedBound

		while (reader.peek() != JsonToken.END_ARRAY) {
			val readObject = resolvedBound.read(reader) ?: throw JsonSyntaxException("List does not accept nulls")
			builder.add(readObject as T)
		}

		reader.endArray()

		return builder.build()
	}
}

class MapAdapter<K, V>(private val boundKey: Class<K>, private val boundValue: Class<V>) : TypeAdapter<Map<K, V>>() {
	private val resolvedKey by lazy {
		resolveBound(boundKey)
	}

	private val resolvedValue by lazy {
		resolveBound(boundValue)
	}

	override fun write(out: JsonWriter, value: Map<K, V>) {
		out.beginArray()

		val resolvedKey = resolvedKey
		val resolvedValue = resolvedValue

		for ((k, v) in value) {
			out.beginArray()
			resolvedKey.write(out, k)
			resolvedValue.write(out, v)
			out.endArray()
		}

		out.endArray()
	}

	override fun read(reader: JsonReader): Map<K, V> {
		reader.beginArray()

		val builder = ImmutableMap.builder<K, V>()

		val resolvedKey = resolvedKey
		val resolvedValue = resolvedValue

		while (reader.peek() != JsonToken.END_ARRAY) {
			reader.beginArray()
			builder.put(resolvedKey.read(reader), resolvedValue.read(reader))
			reader.endArray()
		}

		reader.endArray()

		return builder.build()
	}
}

class StringMapAdapter<V>(private val bound: Class<V>) : TypeAdapter<Map<String, V>>() {
	private val resolvedBound by lazy {
		resolveBound(bound)
	}

	override fun write(out: JsonWriter, value: Map<String, V>) {
		val resolvedBound = resolvedBound

		out.beginObject()

		for ((k, v) in value) {
			out.name(k)
			resolvedBound.write(out, v)
		}

		out.endObject()
	}

	override fun read(reader: JsonReader): Map<String, V> {
		val builder = ImmutableMap.builder<String, V>()

		reader.beginObject()

		val resolvedBound = resolvedBound

		while (reader.peek() != JsonToken.END_OBJECT) {
			builder.put(reader.nextName(), resolvedBound.read(reader))
		}

		reader.endObject()

		return builder.build()
	}
}

private class LazyTypeProvider<T : Any?>(private val bound: Class<T>) : TypeAdapter<T>() {
	private val resolved by lazy { resolveBound(bound) }

	override fun write(out: JsonWriter, value: T) {
		resolved.write(out, value)
	}

	override fun read(`in`: JsonReader): T {
		return resolved.read(`in`)
	}
}

/**
 * TypeAdapter для классов которые создаются единожды и более не меняются ("бетонных классов").
 */
class KConcreteTypeAdapter<T : Any>(
	val bound: KClass<T>,
	val types: ImmutableList<Pair<KProperty1<T, *>, TypeAdapter<*>>>,
	val asList: Boolean = false
) : TypeAdapter<T>() {
	fun register(gsonBuilder: GsonBuilder) {
		gsonBuilder.registerTypeAdapter(bound.java, this)
	}

	private val returnTypeCache = Object2ObjectArrayMap<KProperty1<T, *>, KType>()
	private val mapped = Object2IntArrayMap<String>()

	private val loggedMisses = ObjectArraySet<String>()

	init {
		for ((field) in types) {
			returnTypeCache[field] = field.returnType
		}

		mapped.defaultReturnValue(-1)

		for ((i, pair) in types.withIndex()) {
			mapped[pair.first.name] = i
		}
	}

	private val regularFactory: KFunction<T> = bound.constructors.firstOrNull first@{
		if (it.parameters.size == types.size) {
			val iterator = types.iterator()

			for (param in it.parameters) {
				val nextParam = iterator.next()

				val a = param.type
				val b = nextParam.first.returnType

				if (!a.isSupertypeOf(b) || a.isMarkedNullable != b.isMarkedNullable) {
					return@first false
				}
			}

			return@first true
		}

		return@first false
	} ?: throw NoSuchElementException("Unable to determine constructor for ${bound.qualifiedName} matching (${types.joinToString(", ")})")

	private val syntheticFactory: Constructor<T>? = try {
		bound.java.getDeclaredConstructor(*types.map { (it.first.returnType.classifier as KClass<*>).java }.also {
			it as MutableList
			it.add(Int::class.java)
			it.add(DefaultConstructorMarker::class.java)
		}.toTypedArray())
	} catch(_: NoSuchMethodException) {
		null
	}

	private val syntheticPrimitives: Array<Any?>?

	init {
		if (syntheticFactory == null) {
			syntheticPrimitives = null
		} else {
			syntheticPrimitives = arrayOfNulls(syntheticFactory.parameters.size)

			for ((i, param) in syntheticFactory.parameters.withIndex()) {
				val type = param.parameterizedType as? Class<*> ?: continue

				if (type.isPrimitive) {
					syntheticPrimitives[i] = when (type.name) {
						"boolean" -> false
						"int" -> 0
						"long" -> 0L
						"short" -> (0).toShort()
						"byte" -> (0).toByte()
						"char" -> (0).toChar()
						"float" -> 0f
						"double" -> 0.0
						else -> throw IllegalArgumentException("mamma mia: ${type.name}")
					}
				}
			}
		}
	}

	override fun write(out: JsonWriter, value: T) {
		out.beginObject()

		for ((field, adapter) in types) {
			out.name(field.name)
			@Suppress("unchecked_cast")
			(adapter as TypeAdapter<Any>).write(out, (field as KProperty1<T, Any>).get(value))
		}

		out.endObject()
	}

	override fun read(reader: JsonReader): T {
		if (asList) {
			reader.beginArray()
		} else {
			reader.beginObject()
		}

		val presentValues = BooleanArray(types.size)
		val readValues = arrayOfNulls<Any>(types.size)

		if (asList) {
			val iterator = types.iterator()
			var fieldId = 0

			while (reader.peek() != JsonToken.END_ARRAY) {
				if (!iterator.hasNext()) {
					val name = fieldId.toString()

					if (loggedMisses.add(name)) {
						LOGGER.warn("Skipping JSON field with name $name because ${bound.simpleName} has no such field")
					}

					reader.skipValue()
					fieldId++

					continue
				}

				val (field, adapter) = iterator.next()

				try {
					readValues[fieldId] = adapter.read(reader)
					presentValues[fieldId] = true
				} catch(err: Throwable) {
					throw JsonSyntaxException("Exception reading field ${field.name} for ${bound.qualifiedName}", err)
				}

				fieldId++
			}
		} else {
			while (reader.peek() != JsonToken.END_OBJECT) {
				val name = reader.nextName()
				val fieldId = mapped.getInt(name)

				if (fieldId == -1) {
					if (loggedMisses.add(name)) {
						LOGGER.warn("Skipping JSON field with name $name because ${bound.simpleName} has no such field")
					}

					reader.skipValue()
				} else {
					val (field, adapter) = types[fieldId]

					try {
						readValues[fieldId] = adapter.read(reader)
						presentValues[fieldId] = true
					} catch(err: Throwable) {
						throw JsonSyntaxException("Exception reading field ${field.name} for ${bound.qualifiedName}", err)
					}
				}
			}
		}

		for (i in readValues.indices) {
			if (readValues[i] is String) {
				readValues[i] = (readValues[i] as String).intern()
			}
		}

		if (asList) {
			reader.endArray()
		} else {
			reader.endObject()
		}

		if (syntheticFactory == null || presentValues.all { it }) {
			for ((i, pair) in types.withIndex()) {
				val (field) = pair

				if (readValues[i] == null) {
					if (!returnTypeCache[field]!!.isMarkedNullable) {
						throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} does not accept nulls")
					}

					if (!regularFactory.parameters[i].isOptional && !presentValues[i]) {
						throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} must be defined (even just as null)")
					}
				}
			}

			@Suppress("unchecked_cast")
			return regularFactory.call(*readValues as Array<out Any>)
		} else {
			val ints = if (presentValues.size % 31 == 0) presentValues.size / 31 else presentValues.size / 31 + 1
			val copied = readValues.copyOf(readValues.size + ints + 1)
			var intIndex = readValues.size
			var target = 0
			var targetMove = 0

			for (bool in presentValues) {
				if (!bool) {
					target = target.or(1.shl(targetMove))
				}

				targetMove++

				if (targetMove >= 32) {
					copied[intIndex++] = target
					target = 0
					targetMove = 0
				}
			}

			if (targetMove != 0) {
				copied[intIndex] = target
			}

			val syntheticPrimitives = syntheticPrimitives!!

			for ((i, pair) in types.withIndex()) {
				if (copied[i] != null) {
					continue
				}

				val (field) = pair
				val param = regularFactory.parameters[i]

				if (!param.isOptional && !presentValues[i]) {
					throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} is missing")
				}

				if (returnTypeCache[field]!!.isMarkedNullable) {
					continue
				}

				if (param.isOptional && !presentValues[i]) {
					copied[i] = syntheticPrimitives[i]
					continue
				}

				throw JsonSyntaxException("Field ${field.name} of ${bound.qualifiedName} does not accept nulls")
			}

			return syntheticFactory.newInstance(*copied)
		}
	}

	class Builder<T : Any>(val clazz: KClass<T>) {
		private val types = ArrayList<Pair<KProperty1<T, *>, TypeAdapter<*>>>()

		/**
		 * Добавляет поле с определённым адаптером
		 */
		fun <V> plain(field: KProperty1<T, V>, adapter: TypeAdapter<V>): Builder<T> {
			types.add(field to adapter)
			return this
		}

		/**
		 * Добавляет поле(я) без generic типов и без преобразователей
		 */
		fun plain(vararg fields: KProperty1<T, *>): Builder<T> {
			for (field in fields) {
				val returnType = field.returnType
				val classifier = returnType.classifier as? KClass<*> ?: throw ClassCastException("Unable to cast ${returnType.classifier} to KClass of property ${field.name}!")

				if (classifier.isSuperclassOf(Float::class)) {
					types.add(field to TypeAdapters.FLOAT)
				} else if (classifier.isSuperclassOf(Double::class)) {
					types.add(field to TypeAdapters.DOUBLE)
				} else if (classifier.isSuperclassOf(Int::class)) {
					types.add(field to TypeAdapters.INTEGER)
				} else if (classifier.isSuperclassOf(Long::class)) {
					types.add(field to TypeAdapters.LONG)
				} else if (classifier.isSuperclassOf(String::class)) {
					types.add(field to TypeAdapters.STRING)
				} else if (classifier.isSuperclassOf(Boolean::class)) {
					types.add(field to TypeAdapters.BOOLEAN)
				} else {
					types.add(field to LazyTypeProvider(classifier.java))
				}
			}

			return this
		}

		/**
		 * Добавляет поле(я) без generic типов.
		 *
		 * Если поле с данным именем не найдено, кидается [NoSuchElementException]
		 */
		fun plain(vararg fields: String): Builder<T> {
			val members = clazz.memberProperties

			return this.plain(*Array(fields.size) {
				val field = fields[it]
				return@Array members.firstOrNull { it.name == field } ?: throw NoSuchElementException("Unable to find field $field")
			})
		}

		/**
		 * Добавляет поле, которое содержит список значений V (без null).
		 *
		 * Список неизменяем (создаётся объект [ImmutableList])
		 */
		fun <V : Any> list(field: KProperty1<T, List<V>?>, type: Class<V>): Builder<T> {
			types.add(field to ListAdapter(type))
			return this
		}

		/**
		 * Добавляет поле, которое содержит список значений V (без null).
		 *
		 * Список неизменяем (создаётся объект [ImmutableList])
		 */
		inline fun <reified V : Any> list(field: KProperty1<T, List<V>?>): Builder<T> {
			return this.list(field, V::class.java)
		}

		/**
		 * Добавляет поле-таблицу, которое кодируется как [[key, value], [key, value], ...]
		 *
		 * Таблица неизменяема (создаётся объект [ImmutableMap])
		 */
		fun <K, V> map(field: KProperty1<T, Map<K, V>>, keyType: Class<K>, valueType: Class<V>): Builder<T> {
			types.add(field to MapAdapter(keyType, valueType))
			return this
		}

		/**
		 * Добавляет поле-таблицу, которое кодируется как [[key, value], [key, value], ...]
		 *
		 * Таблица неизменяема (создаётся объект [ImmutableMap])
		 */
		fun <K : Any, V : Any> map(field: KProperty1<T, Map<K, V>>, keyType: KClass<K>, valueType: KClass<V>): Builder<T> {
			types.add(field to MapAdapter(keyType.java, valueType.java))
			return this
		}

		/**
		 * Добавляет поле-таблицу, которое кодируется как {"a": value, "b": value, ...}
		 *
		 * Таблица неизменяема (создаётся объект [ImmutableMap])
		 */
		fun <V> map(field: KProperty1<T, Map<String, V>>, valueType: Class<V>): Builder<T> {
			types.add(field to StringMapAdapter(valueType))
			return this
		}

		/**
		 * Добавляет поле-таблицу, которое кодируется как {"a": value, "b": value, ...}
		 *
		 * Таблица неизменяема (создаётся объект [ImmutableMap])
		 */
		fun <V : Any> map(field: KProperty1<T, Map<String, V>>, valueType: KClass<V>): Builder<T> {
			types.add(field to StringMapAdapter(valueType.java))
			return this
		}

		fun build(asList: Boolean = false): KConcreteTypeAdapter<T> {
			return KConcreteTypeAdapter(clazz, ImmutableList.copyOf(types), asList = asList)
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
